#!/usr/bin/perl -wT
use CGI qw(:standard *table);
use CGI::Carp qw(warningsToBrowser fatalsToBrowser);
use strict;

# tighten up the PATH
$ENV{"PATH"} = "/usr/bin";
print header;
print start_html( -title=>"Compiler Parsing Lab",
		  -style=>{-src=>"../style.css"}), "\n";

print "<div id=&quot;nav&quot;>";
print "<a href = \"http://www.cs.wpi.edu/~kal/courses/compilers/modules.html\">[Modules]</a>";
print "<a href = \"http://www.cs.wpi.edu/~kal/courses/compilers/syllabus.html\">[Syllabus]</a>";
print "<a href = \"http://www.cs.wpi.edu/~kal/courses/compilers/project/index.html\">[Project]</a>";
print "<a href = \"http://users.wpi.edu/~alexvo/index.html\">[Lab Index]</a>";
print "</div>";

print h1("Compiler Parsing Lab"), "\n";
print p, h2("by Todd Cooper &amp; Karen Lemone"), "\n";

my $instruct=<<ENDINSTRUCT;
<p>
In this lab you will use yacc (or its newer version) Bison as well as lex (flex)
to create a parser for a small subset of <i>Javalet</i>
</p>

<p>
The yacc (bison) program is in the left box, the related lex (flex) code is in the bottom box 
and the input program for the generated parser is in the right box.
</p>

<p> 
When you press the button <em>Edit the text and click me</em>, the following 
will happen:<br/>
</p>
<ol>
<li>The yacc/bison code in the left box will be run by bison (yacc)</li>
<li>The lex/flex code will be run by flex (lex)</li>
<li>The C code created by flex will be compiled by  cc, the C compiler</li>
<li>The C code created by bison will be compiled by cc, the C compiler, creating an executable parser</li>
<li>The input in the right box is executed by this parser producing a bottom-up parse<br/>
(The parser prints the left hand side of productions just like you will in Project, Part 2)</li>
</ol>

<p>
The results will be displayed on the bottom of the page.<br/> 
You might have to scroll down in your browser to see them.
</p>

<p>
<b> Step 1 </b> Looking at the bison/yacc code in the left box, write the BNF in standard form. <br/>
<em>Hint: The first production is</em>
</p>

<p>
<code>lines --> epsilon | lines line</code>
</p>

<p>
<b> Step 2 </b> It is really hard to see the bottom-up parse from the output, so change the input to contain just 
the first statement <i> 1 + 1; </i> and press the button again. Now you should be able to 
create a parse from the output (which is the reverse of a leftmost derivation). Show it 
as a parse tree.
</p>

<p>
<b> Step 3 </b> Now,  change the yacc/bison grammar in the left box so that it recognizes division. Note that the lex code 
already recognizes "/", calling the token <code>DIVIDE</code>. (Ask us if you don't see this)<br/>
Run an input to show that the division works (if it does! If not, fix your production) and
show us your output. Draw the parse tree from the output.
</p>

<p>
<b> Step 4 </b> In your written BNF, add productions to perform exponentiation, "^". 
Show it to us before you go on. 
Now add the bison code for ^ to the left box, enter a statement using ^ in the right box and click on the button.
Note that the lex code for "^" is already there with the token called <code>POWER</code>. 
Show us when it works.
</p>

<p>
<b> Step 5</b> Moving to unix/linux:
</p>

<ul>
<li> Create the lex file, say, lex.2</li>
<li> Run lex.2  through lex (Type lex lex.2) to create <i> lex.yy.c </i></li>
<li> Now create the yacc file, say ly.y and  run ly.y through yacc (Type <code>yacc -d ly.y</code>) 
This creates  a file called  <code>yacc.tab.c</code> and the "-d" creates a file of definitions 
called <code>y.tab.h</code>.  When these are compiled they produce our parser. Stop and look at these files
and show them to us.</li>
<li> Now compile these 2 C programs:
<p>
<code>cc  y.tab.c lex.yy.c</code><br/>

which, as usual, creates the file a.out. This is the Parser!
</p>
</li>

<li> For now, we will run the Parser from the Command line.

<p>
Type:
</p>
<p>
<code>
../a.out<br/>
2 + 3 * 4
</code>
</p>

<p>
Show us your output.
</p>
</li>
<li> Run a few more programs.</li>
<li> Now you are ready to do Project, Part 2.</li>
</ul>

<p>
</p>
ENDINSTRUCT
print p, start_form(-method => "post",
		    -action => "parsing_lab2.cgi#output"), "\n";

print $instruct;
my $codefilename = "parsing_lab_$$";
my $inputlexcode = param("lex_code");
my $inputyacccode = param("yacc_code");
my $inputtext = param("input");

open(CODELEXF,">$codefilename.l") 
    or dienice("Couldn't open $codefilename.l for creation: $!");
print CODELEXF $inputlexcode;
close CODELEXF;
open(CODEYACCF,">$codefilename.y") 
    or dienice("Couldn't open $codefilename.y for creation: $!");
print CODEYACCF $inputyacccode;
close CODEYACCF;
open(INPUTF,">$codefilename.input") 
    or dienice("Couldn't open $codefilename.input for creation: $!");
print INPUTF $inputtext;
close INPUTF;

my $flexline = "flex -o$codefilename.lex.c $codefilename.l";
my $yaccline = "bison -d $codefilename.y";
my $compileline1 = "cc -c $codefilename.lex.c";
my $compileline2 = "cc -c $codefilename.y.c";
my $compileline3 = "cc -o$codefilename $codefilename.lex.o $codefilename.y.o -lfl -lm";
my $runline = "./$codefilename < $codefilename.input";
my $output;
if (runandoutput($yaccline))
{
    $output .= "The bison/yacc command returned the error: $?\n";
}
elsif (!myrename("$codefilename.tab.h", "parsing_lab.h"))
{
    $output .= "The mv 1 command returned the error: $?\n";
}
elsif (!myrename("$codefilename.tab.c", "$codefilename.y.c"))
{
    $output .= "The mv 2 command returned the error: $?\n";
}
elsif (runandoutput($flexline))
{
    $output .= "The flex command returned the error: $?\n";
}
elsif (runandoutput($compileline1))
{
    $output .= "The compile 1 command returned the error: $?\n";
}
elsif (runandoutput($compileline2))
{
    $output .= "The compile 2 command returned the error: $?\n";
}
elsif (runandoutput($compileline3))
{
    $output .= "The compile 3 command returned the error: $?\n";
}
elsif (runandoutput($runline))
{
    $output .= "The run command returned the error: $?\n";
}

print start_table();
print Tr(td({colspan => 2},
	    [submit(-name=>'submit',
		    -value=>'Edit the text and click me')])), "\n";
print Tr(td( {width => "50%"}, ["Yacc/Bison Parsing Program"]),
	 td( {width => "50%"}, ["Yacc/Bison Parsing Input"])), "\n";
print Tr(td( {width => "50%"}, textarea(-name => 'yacc_code',
					-default => $inputyacccode,
                                        -rows => 21,
                                        -columns => 42)),
         td( {width => "50%"}, textarea(-name => 'input',
					-default => $inputtext,
                                        -rows => 21,
                                        -columns => 42))), "\n";
print Tr(td( {width => "50%"}, ["Lex Scanner Program"])), "\n";
print Tr(td( {width => "50%"}, textarea(-name => 'lex_code',
					-default => $inputlexcode,
                                        -rows => 21,
                                        -columns => 42))), "\n";

my $thtext = "Edit the program and input above" . br();
$thtext .= "and click on the button to see the result.";
print Tr(th( {align => "left"}, [$thtext] ));

print end_table(), "\n";
my $restext = "The result output or running the input on the parsing program";
$restext .= " is below.";

print "<a name=\"output\"></a>";
print p($restext), "\n";
print "<fieldset><legend>OUTPUT</legend>";
print p($output), "\n";
print "</fieldset>";
print endform(), "\n";

my $htmlstrict=<<ENDHTMLSTRICT;
<p>
    <a href="http://validator.w3.org/check?uri=referer"><img
        src="http://www.w3.org/Icons/valid-xhtml10-blue"
        alt="Valid XHTML 1.0 Transitional" height="31" width="88" /></a>
  </p>
ENDHTMLSTRICT
print $htmlstrict;
print end_html;

# cleanup files we made
unlink("$codefilename.l");
unlink("$codefilename.y");
unlink("$codefilename.input");
unlink("$codefilename.tab.h");
unlink("$codefilename.tab.c");
unlink("$codefilename.lex.c");
unlink("$codefilename.y.c");
unlink("$codefilename.lex.o");
unlink("$codefilename.y.o");
unlink("$codefilename");
unlink("parsing_lab.h");

sub runandoutput
{
    my($cmd) = @_;
    $output .= strong($cmd) . br();
    $output .= "\n";

    open(SUBPROC, "$cmd 2>&1 |")
	or dienice("Couldn't open pipe from $cmd: $!");
	
    foreach my $cmdret (<SUBPROC>)
    {
	chomp $cmdret;
	$output .= code($cmdret) . br() . "\n";
    }
    close SUBPROC;
    return(($? >> 8) & 255);
}

sub myrename
{
    my($arg1, $arg2) = @_;
    $output .= strong("mv $arg1 $arg2") . br();
    $output .= "\n";
    return(rename($arg1, $arg2));
}

sub dienice 
{
    my($msg) = @_;
    print "<h2>Error</h2>\n";
    print $msg;
    exit;
}
