#!/usr/bin/perl -wT
use CGI qw(:standard *table);
use CGI::Carp qw(warningsToBrowser fatalsToBrowser);
use strict;

# tighten up the PATH
$ENV{"PATH"} = "/usr/bin";
print header;
print start_html( -title=>"Compiler Lab #3 - Semantics",
		  -style=>{-src=>"../style.css"}), "\n";

print "<div id=&quot;nav&quot;>";
print "<a href = \"http://www.cs.wpi.edu/~kal/courses/compilers/modules.html\">[Modules]</a>";
print "<a href = \"http://www.cs.wpi.edu/~kal/courses/compilers/syllabus.html\">[Syllabus]</a>";
print "<a href = \"http://www.cs.wpi.edu/~kal/courses/compilers/project/index.html\">[Project]</a>";
print "<a href = \"http://users.wpi.edu/~alexvo/index.html\">[Lab Index]</a>";
print "</div>";

print h1("Semantics lab for the compiler course"), "\n";
print h2("by Todd Cooper and Karen Lemone"), "\n";

my $instruct=<<ENDINSTRUCT;
<p>
In this lab you will add semantic actions to create an abstract syntax tree.
</p><p>
The yacc (bison) program is in the left box and input to that program is in 
the right box.  The lex code is the bottom box.
</p>
<p> 
When you press the button <em>Edit the text and click me</em>, the following 
will happen.
</p>
<ol>
<li>The web page will be submitted to bison (yacc) and flex (lex), 
to create a C program.</li>
<li>This C program is compiled by the cc compiler.</li>
<li>The resultant executable program is run on the input.</li>
<li>Like the other labs, the results will be displayed on the bottom 
of the page.
(You might have to scroll down in your browser to see them.)</li>
</ol>


<b> Step 1 </b> Edit the input, removing all but the first statement, 1 + 1;. 
Click on the button and examine the output. Now (on paper), show a trace of how
the bottom-up parse creates the abstract syntax tree:</p><p>

<img src="../treeforlab.jpg" alt="Tree for Lab" /> 
 <br />
The <i>printtree</i> function prints this as: <i> (+ 1 1) </i></p><p>

<b> Step 2 </b>  Change the yacc/bison grammar so that it recognizes division  and add that node to the AST tree.<br />
Run an expression as input to show that the division works.
</p><p>

<b> Step 3 </b>
Add the <code>pow()</code> function (^) to compute the
power of two numbers; include both the syntax (the yacc productions) and semantics (the \$\$ = stuff)<br />
Again, run an example to show power works.<br /></p><p>

<b> Step 4 </b> Now read (yes, read it first!) and begin Project Part 3. Have fun! 
</p>
ENDINSTRUCT
print p, start_form(-method => "post",
		    -action => "semantics_lab.cgi#output"), "\n";

print $instruct;
my $codefilename = "parsing_lab_$$";
my $inputlexcode = param("lex_code");
my $inputyacccode = param("yacc_code");
my $inputtext = param("input");

open(CODELEXF,">$codefilename.l") 
    or dienice("Couldn't open $codefilename.l for creation: $!");
print CODELEXF $inputlexcode;
close CODELEXF;
open(CODEYACCF,">$codefilename.y") 
    or dienice("Couldn't open $codefilename.y for creation: $!");
print CODEYACCF $inputyacccode;
close CODEYACCF;
open(INPUTF,">$codefilename.input") 
    or dienice("Couldn't open $codefilename.input for creation: $!");
print INPUTF $inputtext;
close INPUTF;

my $flexline = "flex -o$codefilename.lex.c $codefilename.l";
my $yaccline = "bison -d $codefilename.y";
my $compileline1 = "cc -g -c $codefilename.lex.c";
my $compileline2 = "cc -g -c $codefilename.y.c";
my $compileline3 = "cc -g -o$codefilename $codefilename.lex.o $codefilename.y.o -lfl -lm";
my $runline = "./$codefilename < $codefilename.input";
my $output;
if (runandoutput($yaccline))
{
    $output .= "The bison/yacc command returned the error: $?\n";
}
elsif (!myrename("$codefilename.tab.h", "parsing_lab.h"))
{
    $output .= "The mv 1 command returned the error: $?\n";
}
elsif (!myrename("$codefilename.tab.c", "$codefilename.y.c"))
{
    $output .= "The mv 2 command returned the error: $?\n";
}
elsif (runandoutput($flexline))
{
    $output .= "The flex command returned the error: $?\n";
}
elsif (runandoutput($compileline1))
{
    $output .= "The compile 1 command returned the error: $?\n";
}
elsif (runandoutput($compileline2))
{
    $output .= "The compile 2 command returned the error: $?\n";
}
elsif (runandoutput($compileline3))
{
    $output .= "The compile 3 command returned the error: $?\n";
}
elsif (runandoutput($runline))
{
    $output .= "The run command returned the error: $?\n";
}

print start_table();
print Tr(td({colspan => 2},
	    [submit(-name=>'submit',
		    -value=>'Edit the text and click me')])), "\n";
print Tr(td( {width => "50%"}, ["Yacc/Bison Parsing Program"]),
	 td( {width => "50%"}, ["Yacc/Bison Parsing Input"])), "\n";
print Tr(td( {width => "50%"}, textarea(-name => 'yacc_code',
					-default => $inputyacccode,
                                        -rows => 21,
                                        -columns => 42)),
         td( {width => "50%"}, textarea(-name => 'input',
					-default => $inputtext,
                                        -rows => 21,
                                        -columns => 42))), "\n";
print Tr(td( {width => "50%"}, ["Lex Scanner Program"])), "\n";
print Tr(td( {width => "50%"}, textarea(-name => 'lex_code',
					-default => $inputlexcode,
                                        -rows => 21,
                                        -columns => 42))), "\n";

my $thtext = "Edit the program and input above" . br();
$thtext .= "and click on the button to see the result.";
print Tr(th( {align => "left"}, [$thtext] ));

print end_table(), "\n";
my $restext = "The result output or running the input on the parsing program";
$restext .= " is below.";

print "<a name=\"output\"></a>";
print p($restext), "\n";
print "<fieldset><legend>OUTPUT</legend>";
print p($output), "\n";
print "</fieldset>";

print endform(), "\n";

my $htmlstrict=<<ENDHTMLSTRICT;
<p>
    <a href="http://validator.w3.org/check?uri=referer"><img
        src="http://www.w3.org/Icons/valid-xhtml10-blue"
        alt="Valid XHTML 1.0 Transitional" height="31" width="88" /></a>
  </p>
ENDHTMLSTRICT
print $htmlstrict;
print end_html;

# cleanup files we made
unlink("$codefilename.l");
unlink("$codefilename.y");
unlink("$codefilename.input");
unlink("$codefilename.tab.h");
unlink("$codefilename.tab.c");
unlink("$codefilename.lex.c");
unlink("$codefilename.y.c");
unlink("$codefilename.lex.o");
unlink("$codefilename.y.o");
unlink("$codefilename");
unlink("parsing_lab.h");

sub runandoutput
{
    my($cmd) = @_;
    $output .= strong($cmd) . br();
    $output .= "\n";

    open(SUBPROC, "$cmd 2>&1 |")
	or dienice("Couldn't open pipe from $cmd: $!");
	
    foreach my $cmdret (<SUBPROC>)
    {
	$output .= code($cmdret) . br() . "\n";
    }
    close SUBPROC;
    return(($? >> 8) & 255);
}

sub myrename
{
    my($arg1, $arg2) = @_;
    $output .= strong("mv $arg1 $arg2") . br();
    $output .= "\n";
    return(rename($arg1, $arg2));
}

sub dienice 
{
    my($msg) = @_;
    print "<h2>Error</h2>\n";
    print $msg;
    exit;
}
